
#include <stdio.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SimpleMenu.h>

#ifdef HAVE_XPM
#include <X11/xpm.h>
#endif

#include "renju.h"
#include "xrenju.h"
#include "prototype.h"

#define APP_CLASS	"XRenju"

#ifdef HAVE_XPM
#include "pixmap/black.xpm"
#include "pixmap/white.xpm"
#include "pixmap/grid.xpm"
#include "pixmap/centergrid.xpm"
#include "pixmap/large.black.xpm"
#include "pixmap/large.white.xpm"
#include "pixmap/large.grid.xpm"
#include "pixmap/large.centergrid.xpm"
#else
#include "bitmap/black.xbm"
#include "bitmap/white.xbm"
#include "bitmap/grid.xbm"
#include "bitmap/centergrid.xbm"
#include "bitmap/large.black.xbm"
#include "bitmap/large.white.xbm"
#include "bitmap/large.grid.xbm"
#include "bitmap/large.centergrid.xbm"
#endif

Display *disp;
int screen;

Pixel KomaPix[K_MAX], KomaMaskPix[K_MAX];
int KomaWidth, KomaHeight;

board_t bod;
int n;
int putn;
int player[2];
FILE *CmdIn[2];
FILE *CmdOut[2];
char *ExtCom[2];
XtInputId ExtComID[2];
int Mode;
RenjuCmd history[BOARDSIZE*BOARDSIZE];

#ifdef HAVE_XPM
XpmAttributes attr;
#endif

int depth;
unsigned long Blackpixel, Whitepixel;
GC blackGC, whiteGC;

XtAppContext app_con;

Widget rootW;
Widget formW;
Widget buttonsW;
Widget configW;

Widget initW;
Widget startW;
Widget saveW;
Widget loadW;
Widget stepLoadW;
Widget thinkW;
Widget skipW;
Widget quitW;
Widget messageW;
Widget boardW;
Widget player1W;
Widget player2W;
Widget player1ListW;
Widget player2ListW;

Widget notifyW;

char *ordStr[2];
char *histFile;
Boolean notify;
Boolean large;

static XtActionsRec actions[] = {
    {"QuitProc", (XtActionProc) QuitProc},
    {"RedrawBoardProc", (XtActionProc) RedrawBoardProc},
    {"PutKoma", (XtActionProc) PutKoma},
    {"PutKomaSafe", (XtActionProc) PutKomaSafe},
};

static String fallback_resources[] = {
    "*.translations: #override\\n\
            Shift<Key>Q:	QuitProc()",
    "*.boardW.translations: #override\\n\
            <Expose>:	RedrawBoardProc()\\n\
            <Btn1Down>:	PutKomaSafe()\\n\
            <Btn3Down>:	PutKoma()",

    "*.buttonsW.fromHoriz:	boardW",
    "*.buttonsW*showGrip:	False",
    "*.configW.fromHoriz:	boardW",
    "*.configW.fromVert:	buttonsW",
    "*.configW*showGrip:	False",
    "*.messageW.fromVert:	boardW",
    "*.messageW.height:		100",
    "*.messageW.width:		10",
    "*.messageW.scrollVertical:		WhenNeeded",
    "*.messageW.scrollHorizontal:	WhenNeeded",

    "*.player1W.menuName:	 player1ListW",
    "*.player2W.menuName:	 player2ListW",

    "*.quitW.label:	Quit",
    "*.initW.label:	Initialize",
    "*.startW.label:	Start",
    "*.saveW.label:	Save",
    "*.loadW.label:	Load",
    "*.stepLoadW.label:	Step Load",
    "*.thinkW.label:	Think",
    "*.skipW.label:	Pass",

    "*.notifyW.label:	Notify",

#if 0
    "*.international:	True",
    "*.FontSet:		-*--14-*",
#endif
    NULL
};

static XrmOptionDescRec options[] = {
    {"-black", ".Black", XrmoptionSepArg, NULL},
    {"-extblack", ".ExtBlack", XrmoptionSepArg, NULL},
    {"-white", ".White", XrmoptionSepArg, NULL},
    {"-extwhite", ".ExtWhite", XrmoptionSepArg, NULL},
    {"-histfile", ".HistFile", XrmoptionSepArg, NULL},
    {"-notify",    "Notify",  XrmoptionNoArg,  "TRUE"},
    {"-nonotify",  "Notify",  XrmoptionNoArg,  "FALSE"},
    {"-large",    "Large",  XrmoptionNoArg,  "TRUE"},
    {"-normal",    "Large",  XrmoptionNoArg,  "FALSE"},
};

struct app_resources_s {
    String black;
    String extblack;
    String blacklabel;
    String white;
    String extwhite;
    String whitelabel;
    String histfile;
    String humanlabel;
    String comlabel;
    String extlabel;
    Boolean notify;
    Boolean large;
} app_resources;

#define xoffset(entry)	XtOffset(struct app_resources_s *, entry)
#define reslist(name, class, entry, default) \
		{name, class, XtRString, sizeof(String), \
		    xoffset(entry), XtRString, default}
static XtResource resources[] = {
    reslist("black", "Black", black, "HUMAN"),
    reslist("extblack", "ExtBlack", extblack, "think_sample"),
    reslist("blacklabel", "BlackLabel", blacklabel, "Black"),
    reslist("white", "White", white, "COMPUTER"),
    reslist("extwhite", "ExtWhite", extwhite, "think_sample"),
    reslist("whitelabel", "WhiteLabel", whitelabel, "White"),
    reslist("histfile", "HistFile", histfile, "renju.hst"),
    reslist("humanlabel", "HumanLabel", humanlabel, "Human"),
    reslist("comlabel", "ComLabel", comlabel, "Computer"),
    reslist("extlabel", "ExtLabel", extlabel, "External Computer"),
    {"notify", "Notify", XtRBoolean, sizeof(Boolean), xoffset(notify),
	XtRImmediate, (XtPointer) True},
    {"large", "Large", XtRBoolean, sizeof(Boolean), xoffset(large),
	XtRImmediate, (XtPointer) False},
};

typedef struct {
    Widget *widget;
    char *name;
    XtCallbackProc proc;
} ButtonList;

static ButtonList RenjuButton[] = {
    {&quitW, "quitW", (XtCallbackProc) QuitProc},
    {&initW, "initW", (XtCallbackProc) InitBoard},
    {&startW, "startW", (XtCallbackProc) StartProc},
    {&saveW, "saveW", (XtCallbackProc) SaveProc},
    {&loadW, "loadW", (XtCallbackProc) LoadProc},
    {&stepLoadW, "stepLoadW", (XtCallbackProc) StepLoadProc},
    {&thinkW, "thinkW", (XtCallbackProc) ThinkProc},
    {&skipW, "skipW", (XtCallbackProc) SkipProc},
    {NULL, NULL, NULL}
};

static ButtonList RenjuConfig[] = {
    {&notifyW, "notifyW", (XtCallbackProc) NotifyProc},
    {NULL, NULL, NULL}
};

char *PlayerStr[3];

static void SetButtons(Widget parent, ButtonList *BtnList)
{
    ButtonList *btn;

    btn = BtnList;
    while (btn->widget) {
	*(btn->widget) = XtVaCreateManagedWidget(btn->name,
		    commandWidgetClass, parent, NULL);
	XtAddCallback(*(btn->widget), XtNcallback, btn->proc, NULL);
	btn++;
    }
}

int main(int argc, char *argv[])
{
    int i;
    Widget entry;
    char buf[1024];

    clear_board(bod);

    CmdIn[K_BLACK] = CmdIn[K_WHITE] = NULL;
    CmdOut[K_BLACK] = CmdOut[K_WHITE] = NULL;

    Mode = M_INITIALIZED;
    initJukei();

#ifdef USE_PUTENV
    sprintf(buf, "LANG=%s", DEFAULT_LANG);
    putenv(buf);
#endif
    XtSetLanguageProc(NULL, NULL, NULL);
    rootW = XtVaAppInitialize(&app_con, APP_CLASS, 
		options, XtNumber(options),
		&argc, argv,
		fallback_resources, NULL);
    XtAppAddActions(app_con, actions, XtNumber(actions));
    XtVaGetApplicationResources(rootW, &app_resources,
		resources, XtNumber(resources), NULL);

    switch (app_resources.black[0]) {
    case 'H':
    case 'h':
	player[K_BLACK] = HUMAN;
	break;
    case 'C':
    case 'c':
	player[K_BLACK] = COMPUTER;
	break;
    case 'E':
    case 'e':
	player[K_BLACK] = EXT_COM;
	break;
    default:
	player[K_BLACK] = HUMAN;
	break;
    }
    ExtCom[K_BLACK] = app_resources.extblack;

    switch (app_resources.white[0]) {
    case 'H':
    case 'h':
	player[K_WHITE] = HUMAN;
	break;
    case 'C':
    case 'c':
	player[K_WHITE] = COMPUTER;
	break;
    case 'E':
    case 'e':
	player[K_WHITE] = EXT_COM;
	break;
    default:
	player[K_WHITE] = COMPUTER;
	break;
    }
    ExtCom[K_WHITE] = app_resources.extwhite;

    ordStr[K_BLACK] = app_resources.blacklabel;
    ordStr[K_WHITE] = app_resources.whitelabel;
    histFile = app_resources.histfile;
    notify = app_resources.notify;
    large = app_resources.large;

    PlayerStr[HUMAN] = app_resources.humanlabel;
    PlayerStr[COMPUTER] = app_resources.comlabel;
    PlayerStr[EXT_COM] = app_resources.extlabel;

    getMessage();

    formW = XtVaCreateManagedWidget("formW", formWidgetClass, rootW, NULL);
    boardW = XtVaCreateManagedWidget("boardW", widgetClass, formW, NULL);
    buttonsW = XtVaCreateManagedWidget("buttonsW", panedWidgetClass,
					formW, NULL);
    configW = XtVaCreateManagedWidget("configW", panedWidgetClass, formW, NULL);
    messageW = XtVaCreateManagedWidget("messageW", asciiTextWidgetClass,
					formW, NULL);

    disp = XtDisplay(rootW);
    screen = DefaultScreen(disp);

    depth = DefaultDepthOfScreen(XtScreen(rootW));
    Blackpixel = BlackPixel(disp, screen);
    Whitepixel = WhitePixel(disp, screen);
    blackGC = XCreateGC(disp, RootWindow(disp, 0), 0, NULL);
    XSetForeground(disp, blackGC, Blackpixel);
    whiteGC = XCreateGC(disp, RootWindow(disp, 0), 0, NULL);
    XSetForeground(disp, whiteGC, Whitepixel);

    SetButtons(buttonsW, RenjuButton);
    SetButtons(configW, RenjuConfig);

    player1W = XtVaCreateManagedWidget("player1W",
			menuButtonWidgetClass, buttonsW, NULL);
    player1ListW = XtVaCreatePopupShell("player1ListW",
			simpleMenuWidgetClass, player1W, NULL);
    for (i = 0; i < XtNumber(PlayerStr); i++) {
	entry = XtVaCreateManagedWidget(PlayerStr[i],
			smeBSBObjectClass, player1ListW, NULL);
	XtAddCallback(entry, XtNcallback,
			(XtCallbackProc) Player1Select, (XtPointer) i);
    }
    SetPlayerLabel(player1W, K_BLACK, player[K_BLACK]);

    player2W = XtVaCreateManagedWidget("player2W",
			menuButtonWidgetClass, buttonsW, NULL);
    player2ListW = XtVaCreatePopupShell("player2ListW",
			simpleMenuWidgetClass, player2W, NULL);
    for (i = 0; i < XtNumber(PlayerStr); i++) {
	entry = XtVaCreateManagedWidget(PlayerStr[i],
			smeBSBObjectClass, player2ListW, NULL);
	XtAddCallback(entry, XtNcallback,
			(XtCallbackProc) Player2Select, (XtPointer) i);
    }
    SetPlayerLabel(player2W, K_WHITE, player[K_WHITE]);

    for (i = 0; i < K_MAX; i++) {
	KomaPix[i] = None;
	KomaMaskPix[i] = None;
    }
#ifdef HAVE_XPM
    attr.valuemask = XpmReturnPixels;
    if (large) {
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			large_black_xpm,
			&KomaPix[K_BLACK], &KomaMaskPix[K_BLACK], &attr);
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			large_white_xpm,
			&KomaPix[K_WHITE], &KomaMaskPix[K_WHITE], &attr);
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			large_centergrid_xpm,
			&KomaPix[K_CENTER], &KomaMaskPix[K_CENTER], &attr);
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			large_grid_xpm,
			&KomaPix[K_NOTHING], &KomaMaskPix[K_NOTHING], &attr);
    } else {
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			black_xpm,
			&KomaPix[K_BLACK], &KomaMaskPix[K_BLACK], &attr);
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			white_xpm,
			&KomaPix[K_WHITE], &KomaMaskPix[K_WHITE], &attr);
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			centergrid_xpm,
			&KomaPix[K_CENTER], &KomaMaskPix[K_CENTER], &attr);
	XpmCreatePixmapFromData(disp, RootWindow(disp, 0),
			grid_xpm,
			&KomaPix[K_NOTHING], &KomaMaskPix[K_NOTHING], &attr);
    }
    KomaPix[K_CHECK] = KomaPix[K_NOTHING];
    KomaMaskPix[K_CHECK] = KomaMaskPix[K_NOTHING];
    KomaWidth = attr.width;
    KomaHeight = attr.height;
#else
    if (large) {
	KomaWidth = large_black_width;
	KomaHeight = large_black_height;
	KomaPix[K_BLACK] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)large_black_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
	KomaPix[K_WHITE] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)large_white_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
	KomaPix[K_NOTHING] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)large_grid_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
	KomaPix[K_CENTER] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)large_centergrid_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
    } else {
	KomaWidth = black_width;
	KomaHeight = black_height;
	KomaPix[K_BLACK] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)black_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
	KomaPix[K_WHITE] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)white_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
	KomaPix[K_NOTHING] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)grid_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
	KomaPix[K_CENTER] = XCreatePixmapFromBitmapData(disp,
		    RootWindowOfScreen(XtScreen(rootW)),
		    (char *)centergrid_bits, KomaWidth, KomaHeight,
		    Blackpixel, Whitepixel, depth);
    }
    KomaPix[K_CHECK] = KomaPix[K_NOTHING];
#endif

    XtVaSetValues(boardW, XtNwidth, KomaWidth*BOARDSIZE,
			    XtNheight, KomaHeight*BOARDSIZE, NULL);
    XtVaSetValues(messageW, XtNwidth, KomaWidth*BOARDSIZE, NULL);

    PlayerSelect(player1W, K_BLACK, player[K_BLACK]);
    PlayerSelect(player2W, K_WHITE, player[K_WHITE]);

    XtRealizeWidget(rootW);
    XtAppMainLoop(app_con);

    return 0;
}
