
#include <stdio.h>
#include <stdarg.h>

#include <X11/Xlib.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>

#include "renju.h"
#include "xrenju.h"
#include "prototype.h"

/* $B2hLL$N=q49$($G$9!#(B */
void RedrawBoardProc(Widget w)
{
    int i, j;
    for (i = 0; i < BOARDSIZE; i++) {
	for (j = 0; j < BOARDSIZE; j++) {
	XCopyArea(XtDisplay(rootW), KomaPix[BAN(bod, j, i).koma],
		    XtWindow(w), blackGC, 0, 0,
		    KomaWidth, KomaHeight, KomaWidth*j, KomaHeight*i);
	}
    }
}

/* $B=*N;=hM}$G$9!#(B */
void QuitProc(void)
{
    int i;
    for (i = 0; i < K_MAX; i++) {
	if (KomaPix[i] != None)
	    XFreePixmap(disp, KomaPix[i]);
	if (KomaMaskPix[i] != None)
	    XFreePixmap(disp, KomaMaskPix[i]);
    }
    exit(0);
}

/* $B%a%C%;!<%8$r=PNO$9$kHFMQ4X?t$G$9!#(B */
void PutMessage(char *format, ...)
{
    char str[1024];
    XawTextBlock textblock;
    XawTextPosition textpos;
    va_list args;

    va_start(args, format);
    vsprintf(str, format, args);
    va_end(args);

    textpos = XawTextGetInsertionPoint(messageW);

    textblock.ptr = str;
    textblock.length = strlen(textblock.ptr);
    textblock.firstPos = 0;
    textblock.format = FMT8BIT;

    XtVaSetValues(messageW, XtNeditType, XawtextAppend, NULL);
    XawTextReplace(messageW, textpos, textpos, &textblock);
    XtVaSetValues(messageW, XtNeditType, XawtextRead, NULL);
}

void ClearMessage(void)
{
    XawTextBlock block;
    XawTextPosition textpos;

#if 1
    PutMessage("\n");
#endif

    textpos = XawTextGetInsertionPoint(messageW);

    block.ptr = "";
    block.length = 0;
    block.firstPos = 0;
    block.format = FMT8BIT;

    XtVaSetValues(messageW, XtNeditType, XawtextAppend, NULL);
    XawTextReplace(messageW, 0, textpos, &block);
    XtVaSetValues(messageW, XtNeditType, XawtextRead, NULL);
}

/* $B2hLL$d30It%k!<%F%#%s$b9M$($?6p$NA`:n4X?t$G$9!#(B */
void PutKomaCmd(Widget w, board_t bod, RenjuCmd cmd)
{
    int x, y;
    int i;
    char buf[1024];
    char *str;

    if (cmd.type != CMD_PUT && cmd.type != CMD_SKIP)
	return;

    x = cmd.x;
    y = cmd.y;
    if (cmd.type == CMD_PUT)
	put_koma(bod, n, x, y);

    if (Mode == M_PLAYING && notify) {
	if (cmd.type == CMD_PUT) {
	    posMessage(buf, x, y);
	    str = messages[(int)BAN(bod, x, y).lose[n&1]];
	    if (strcmp(str, "")) {
		PutMessage("%s: %s: %s\n", ordStr[n&1], buf, str);
	    } else {
		PutMessage("%s: %s\n", ordStr[n&1], buf);
	    }
	} else {
	    WriteCmdStr(buf, cmd);
	    i = strlen(buf);
	    if (buf[i-1] == '\n')
		buf[i-1] = '\0';
	    PutMessage("%s: %s\n", ordStr[n&1], buf);
	}
    }

    if (cmd.type == CMD_PUT) {
	XCopyArea(XtDisplay(rootW), KomaPix[BAN(bod, x, y).koma],
		XtWindow(boardW), blackGC, 0, 0,
		KomaWidth, KomaHeight, KomaWidth*x, KomaHeight*y);
	if (BAN(bod, x, y).lose[n&1] == T_WIN ||
	    BAN(bod, x, y).lose[n&1] == T_LOSE ||
	    putn == BOARDSIZE*BOARDSIZE - 1)
	    Mode = M_END;
    }

    history[n] = cmd;
    if (Mode == M_END) {
	if (putn == BOARDSIZE*BOARDSIZE - 1) {
	    PutMessage("%s\n", drawgame);
	} else if (BAN(bod, x, y).lose[n & 1] == T_WIN) {
	    PutMessage("%s\n", messageWon[n & 1]);
	} else if (BAN(bod, x, y).lose[n & 1] == T_LOSE) {
	    PutMessage("%s\n", messageWon[1 - (n & 1)]);
	}
    }
    XFlush(disp);
    for (i = 0; i < 2; i++) {
	if (player[i] == EXT_COM) {
	    WriteCmdFile(CmdOut[i], cmd);
	}
    }

    n++;
    putn++;
    if (Mode == M_PLAYING) {
	switch (player[n & 1]) {
	case COMPUTER:
	case EXT_COM:
	    ThinkCom();
	    break;
	}
    }
}

/* $B2hLL$G%\%?%s$,2!$5$l$?;~$K8F$S=P$5$l$k4X?t$G$9!#(B */
void PutKoma(Widget w, XEvent *ev)
{
    int x, y;
    RenjuCmd cmd;

    if (player[n&1] != HUMAN || Mode == M_END)
	return;

    x = ev->xbutton.x/KomaWidth;
    y = ev->xbutton.y/KomaHeight;

    cmd.type = CMD_PUT;
    cmd.x = x;
    cmd.y = y;

    Mode = M_PLAYING;
    if (BAN(bod, x, y).koma >= K_NOTHING) {
	PutKomaCmd(w, bod, cmd);
    }
}

/* $B2hLL$G%\%?%s$,2!$5$l$?;~$K8F$S=P$5$l$k4X?t$G$9!#(B */
void PutKomaSafe(Widget w, XEvent *ev)
{
    int x, y;
    RenjuCmd cmd;

    if (player[n&1] != HUMAN || Mode == M_END)
	return;

    x = ev->xbutton.x/KomaWidth;
    y = ev->xbutton.y/KomaHeight;

    if (BAN(bod, x, y).lose[n&1] == T_LOSE) {
	PutMessage("%s\n", warning);
	return;
    }

    cmd.type = CMD_PUT;
    cmd.x = x;
    cmd.y = y;

    Mode = M_PLAYING;
    if (BAN(bod, x, y).koma >= K_NOTHING) {
	PutKomaCmd(w, bod, cmd);
    }
}

/* Start$B%\%?%s$r2!$7$?;~$K8F$S=P$5$l$^$9!#(B */
void StartProc(void)
{
    if (Mode == M_END)
	return;

    switch (player[n&1]) {
    case COMPUTER:
    case EXT_COM:
	Mode = M_PLAYING;
	ThinkCom();
	break;
    }
}

/* Init$B%\%?%s$r2!$7$?;~$K8F$S=P$5$l$^$9!#(B */
void InitBoard(void)
{
    int i;
    RenjuCmd cmd;

    if (Mode == M_INITIALIZED)
	return;

    clear_board(bod);
    n = 0;
    putn = 0;
    Mode = M_INITIALIZED;
    RedrawBoardProc(boardW);

    cmd.type = CMD_INIT;
    for (i = 0; i < 2; i++) {
	if (player[i] == EXT_COM)
	    WriteCmdFile(CmdOut[i], cmd);
    }

    ClearMessage();
#if 0
    PutMessage("Initialized.\n");
#endif
}

/* $B30It;W9M%k!<%F%#%s$+$i7k2L$,5"$C$F$-$?;~$K8F$S=P$5$l$^$9!#(B */
void PutExtThinker(int num, int *fd, XtInputId* id)
{
    RenjuCmd cmd;

    if ((n & 1) == num) {
	cmd = ReadCmdFile(CmdIn[num]);
	if (Mode != M_END)
	    PutKomaCmd(boardW, bod, cmd);
    } else if (Mode == M_PLAYING) {
	fprintf(stderr, "%d, %d != %d\n", num, n & 1, n);
	cmd = ReadCmdFile(CmdIn[num]);
    } else {
	cmd = ReadCmdFile(CmdIn[num]);
    }
}

/* $B%3%s%T%e!<%?$N;W9M%k!<%F%#%s$r8F$S=P$7$^$9!#(B */
void ThinkCom(void)
{
    RenjuCmd cmd;

    if (Mode == M_END)
	return;

    switch (player[n & 1]) {
    case COMPUTER:
	cmd = think(bod, n);
	PutKomaCmd(boardW, bod, cmd);
	break;
	
    case EXT_COM:
	cmd.type = CMD_THINK;
	WriteCmdFile(CmdOut[n & 1], cmd);
	break;
    }
}

/* Think$B%\%?%s$,2!$5$l$?;~$K8F$S=P$5$l$^$9!#(B */
void ThinkProc(void)
{
    RenjuCmd cmd;

    if (Mode == M_END)
	return;

    switch (player[n & 1]) {
    case HUMAN:
	Mode = M_PLAYING;
	cmd = think(bod, n);
	PutKomaCmd(boardW, bod, cmd);
	break;
    }
}

/* $B%Q%9MQ$N4X?t(B */
void SkipProc(void)
{
    RenjuCmd cmd;

    if (Mode == M_END)
	return;

    switch(player[n & 1]) {
    case HUMAN:
	Mode = M_PLAYING;
	cmd.type = CMD_SKIP;
	PutKomaCmd(boardW, bod, cmd);
	break;
    }
}

/* $B%W%l%$%d!<$rJQ99$7$?;~$K%i%Y%k$rJQ99$7$^$9!#(B */
void SetPlayerLabel(Widget w, int num, int type)
{
    char buf[1024];

    sprintf(buf, "%s: %s", ordStr[num], PlayerStr[type]);
    XtVaSetValues(w, XtNlabel, buf, NULL);
}

/* $B%W%l%$%d!<$r@_Dj$9$k4X?t$G$9!#(B */
void PlayerSelect(Widget w, int num, int type)
{
    RenjuCmd cmd;

    if (player[num] == EXT_COM) {
	if (CmdOut[num] != NULL) {
	    cmd.type = CMD_END;
	    WriteCmdFile(CmdOut[num], cmd);
	    fclose(CmdOut[num]);
	    CmdOut[num] = NULL;
	}
	if (CmdIn[num] != NULL) {
	    fclose(CmdIn[num]);
	    CmdIn[num] = NULL;
	}
	if (ExtComID[num] != None) {
	    XtRemoveInput(ExtComID[num]);
	    ExtComID[num] = None;
	}
    }

    player[num] = type;
    if (player[num] == EXT_COM) {
	init_pipe(ExtCom[num], &CmdIn[num], &CmdOut[num]);
	if (CmdIn[num] == NULL)
	    fprintf(stderr, "CmdIn[%d] is NULL\n", num);
	if (CmdOut[num] == NULL)
	    fprintf(stderr, "CmdOut[%d] is NULL\n", num);
	cmd.type = CMD_INIT;
	WriteCmdFile(CmdOut[num], cmd);
	cmd.type = CMD_THINK;
	WriteCmdFile(CmdOut[num], cmd);
	cmd = ReadCmdFile(CmdIn[num]);
	if (cmd.type == CMD_END) {
	    fprintf(stderr, "External Thinker %s has bug?\n", ExtCom[num]);
	    player[num] = COMPUTER;
	} else {
	    ExtComID[num] = XtAppAddInput(app_con, fileno(CmdIn[num]),
				(XtPointer) XtInputReadMask,
				(XtInputCallbackProc) PutExtThinker,
				(XtPointer) num);
	}
    }
    SetPlayerLabel(w, num, player[num]);
}

void Player1Select(Widget w, int type)
{
    if (Mode != M_PLAYING)
	PlayerSelect(player1W, K_BLACK, type);
}

void Player2Select(Widget w, int type)
{
    if (Mode != M_PLAYING)
	PlayerSelect(player2W, K_WHITE, type);
}

/* $B%R%9%H%j!<$r%;!<%V$7$^$9!#(B */
void SaveProc(void)
{
    int f;
    PutMessage("Saving...");
    f = saveHistory(histFile, history, n);
    if (f) {
	PutMessage("\nNo Such file or directory!\n");
    } else {
	PutMessage(" Done\n");
    }
}

/* $B%R%9%H%j!<$r%m!<%I$7$^$9!#(B */
void LoadProc(void)
{
    int hn, i;

    hn = loadHistory(histFile, history);
    InitBoard();
    PutMessage("Loading...");
    Mode = M_LOADING;
    for (i = 0; i < hn; i++) {
	PutKomaCmd(boardW, bod, history[i]);
    }
    PutMessage(" Done\n");

}

/* $B%R%9%H%j!<$+$i0l<j%m!<%I$7$^$9!#(B */
void StepLoadProc(void)
{
    int hn;

    if (Mode != M_STEPLOADING) {
	hn = loadHistory(histFile, history);
	InitBoard();
	Mode = M_STEPLOADING;
    }

    if (history[n].type != CMD_END) {
	PutKomaCmd(boardW, bod, history[n]);
    }
}

void NotifyProc(void)
{
    if (notify)
	notify = False;
    else
	notify = True;
}
